@@ -19,11 +19,13 @@
 #include "pos-completer-manager.h"
 #include "pos-completion-bar.h"
 #include "pos-input-surface.h"
+#include "pos-keypad.h"
 #include "pos-logind-session.h"
 #include "pos-main.h"
 #include "pos-osk-widget.h"
 #include "pos-settings-panel.h"
 #include "pos-shortcuts-bar.h"
+#include "pos-style-manager.h"
 #include "pos-vk-driver.h"
 #include "pos-virtual-keyboard.h"
 #include "pos-vk-driver.h"
@@ -114,14 +116,15 @@ struct _PosInputSurface {
   GtkWidget               *osk_terminal;
   GtkWidget               *emoji_picker;
   GtkWidget               *last_layout;
+  GtkWidget               *keypad;
   PosShortcutsBar         *shortcuts_bar;
   PhoshOskFeatures         osk_features;
+  GdkModifierType          latched_modifiers;
 
   /* TODO: this should be an interface for different keyboard drivers */
   PosVkDriver             *keyboard_driver;
 
-  GtkCssProvider          *css_provider;
-  char                    *theme_name;
+  PosStyleManager         *style_manager;
 
   /* menu popover */
   GtkBox                  *menu_box_layouts;
@@ -174,6 +177,14 @@ on_swipe (GtkGestureSwipe *swipe, double velocity_x, double velocity_y, gpointer
   }
 }
 
+
+static void
+pos_input_surface_unlatch_modifiers (PosInputSurface *self)
+{
+  pos_shortcuts_bar_unlatch_modifiers (self->shortcuts_bar);
+}
+
+
 static void
 on_shortcut_activated (PosInputSurface *self, PosShortcut *shortcut, PosShortcutsBar *bar)
 {
@@ -182,7 +193,20 @@ on_shortcut_activated (PosInputSurface *self, PosShortcut *shortcut, PosShortcut
 
   pos_vk_driver_key_press_gdk (self->keyboard_driver,
                                pos_shortcut_get_key (shortcut),
-                               pos_shortcut_get_modifiers (shortcut));
+                               pos_shortcut_get_modifiers (shortcut) | self->latched_modifiers);
+  pos_input_surface_unlatch_modifiers (self);
+}
+
+
+static void
+on_latched_modifiers_changed (PosInputSurface *self, GParamSpec *pspec, PosShortcutsBar *bar)
+{
+  g_return_if_fail (POS_IS_INPUT_SURFACE (self));
+  g_return_if_fail (POS_IS_SHORTCUTS_BAR (bar));
+
+  self->latched_modifiers = pos_shortcuts_bar_get_latched_modifiers (bar);
+
+  g_debug ("Modifiers: 0x%x", self->latched_modifiers);
 }
 
 
@@ -242,7 +266,7 @@ pos_input_surface_is_completion_mode (PosInputSurface *self)
 {
   GtkWidget *osk_widget;
 
-  if (pos_input_surface_is_completer_active(self) == FALSE)
+  if (pos_input_surface_is_completer_active (self) == FALSE)
     return FALSE;
 
   /* no completion in cursor mode */
@@ -333,6 +357,9 @@ pos_input_surface_submit_current_preedit (PosInputSurface *self)
     return;
 
   preedit = g_strdup (pos_completer_get_preedit (self->completer));
+  if (gm_str_is_null_or_empty (preedit))
+    return;
+
   g_debug ("%s: Submitting %s", __func__, preedit);
   pos_completer_set_preedit (self->completer, NULL);
   pos_input_method_send_preedit (self->input_method, "", 0, 0, FALSE);
@@ -340,41 +367,6 @@ pos_input_surface_submit_current_preedit (PosInputSurface *self)
 }
 
 
-/* Select proper style sheet in case of high contrast */
-static void
-on_gtk_theme_name_changed (PosInputSurface *self, GParamSpec *pspec, GtkSettings *settings)
-{
-  const char *style;
-  g_autofree char *name = NULL;
-
-  g_autoptr (GtkCssProvider) provider = gtk_css_provider_new ();
-
-  g_object_get (settings, "gtk-theme-name", &name, NULL);
-
-  if (g_strcmp0 (self->theme_name, name) == 0)
-    return;
-
-  self->theme_name = g_steal_pointer (&name);
-  g_debug ("GTK theme: %s", self->theme_name);
-
-  if (self->css_provider) {
-    gtk_style_context_remove_provider_for_screen (gdk_screen_get_default (),
-                                                  GTK_STYLE_PROVIDER (self->css_provider));
-  }
-
-  if (g_strcmp0 (self->theme_name, "HighContrast") == 0)
-    style = "/sm/puri/phosh/osk-stub/stylesheet/adwaita-hc-light.css";
-  else
-    style = "/sm/puri/phosh/osk-stub/stylesheet/adwaita-dark.css";
-
-  gtk_css_provider_load_from_resource (provider, style);
-  gtk_style_context_add_provider_for_screen (gdk_screen_get_default (),
-                                             GTK_STYLE_PROVIDER (provider),
-                                             GTK_STYLE_PROVIDER_PRIORITY_APPLICATION);
-  g_set_object (&self->css_provider, provider);
-}
-
-
 static void
 on_osk_key_down (PosInputSurface *self, const char *symbol, GtkWidget *osk_widget)
 {
@@ -386,15 +378,24 @@ on_osk_key_down (PosInputSurface *self, const char *symbol, GtkWidget *osk_widge
 
 
 static void
-on_osk_key_symbol (PosInputSurface *self, const char *symbol, GtkWidget *osk_widget)
+on_osk_key_symbol (PosInputSurface *self, const char *symbol)
 {
   gboolean handled;
 
   g_return_if_fail (POS_IS_INPUT_SURFACE (self));
-  g_return_if_fail (osk_widget == NULL || POS_IS_OSK_WIDGET (osk_widget));
 
   g_debug ("Key: '%s' symbol", symbol);
 
+  /* Latched modifiers, send as virtual-keyboard */
+  if (self->latched_modifiers) {
+    PosKeycodeModifier modifier;
+
+    modifier = pos_vk_driver_convert_modifiers (self->keyboard_driver, self->latched_modifiers);
+    pos_vk_driver_key_down (self->keyboard_driver, symbol, modifier);
+    pos_vk_driver_key_up (self->keyboard_driver, symbol);
+    pos_input_surface_unlatch_modifiers (self);
+    return;
+  }
   /* virtual-keyboard, no input method */
   if (!pos_input_method_get_active (self->input_method)) {
     pos_vk_driver_key_down (self->keyboard_driver, symbol, POS_KEYCODE_MODIFIER_NONE);
@@ -491,18 +492,6 @@ on_osk_popover_hidden (PosInputSurface *self)
 }
 
 
-static void
-clipboard_copy_activated (GSimpleAction *action,
-                          GVariant      *parameter,
-                          gpointer       data)
-{
-  PosInputSurface *self = POS_INPUT_SURFACE (data);
-
-  pos_vk_driver_key_down (self->keyboard_driver, "c", POS_KEYCODE_MODIFIER_CTRL);
-  pos_vk_driver_key_up (self->keyboard_driver, "c");
-}
-
-
 static void
 clipboard_paste_activated (GSimpleAction *action,
                            GVariant      *parameter,
@@ -601,9 +590,36 @@ static void
 on_emoji_picker_delete_last (PosInputSurface *self)
 {
   g_debug ("%s", __func__);
-  on_osk_key_symbol (self, "KEY_BACKSPACE", NULL);
+  on_osk_key_symbol (self, "KEY_BACKSPACE");
 }
 
+/* Keypads */
+
+static void
+on_keypad_symbol_pressed (PosInputSurface *self, const char *symbol, PosKeypad *keypad)
+{
+  g_assert (POS_IS_INPUT_SURFACE (self));
+  g_assert (POS_IS_KEYPAD (keypad));
+
+  if (pos_input_method_get_active (self->input_method)) {
+    pos_input_surface_submit_current_preedit (self);
+    pos_input_method_send_string (self->input_method, symbol, TRUE);
+  } else {
+    pos_vk_driver_key_down (self->keyboard_driver, symbol, POS_KEYCODE_MODIFIER_NONE);
+    pos_vk_driver_key_up (self->keyboard_driver, symbol);
+  }
+
+  pos_input_surface_notify_key_press (self);
+}
+
+
+static void
+on_keypad_done (PosInputSurface *self)
+{
+  hdy_deck_set_visible_child (self->deck, self->last_layout);
+}
+
+
 /* menu button */
 
 
@@ -703,6 +719,9 @@ select_layout_change_state (GSimpleAction *action,
   /* popdown popover right away to avoid flicker when switching layouts */
   gtk_popover_popdown (self->menu_popup);
 
+  /* reset all letched modifiers */
+  pos_input_surface_unlatch_modifiers (self);
+
   g_variant_get (parameter, "&s", &layout);
   g_debug ("Layout '%s' selected", layout);
 
@@ -1077,7 +1096,7 @@ pos_input_surface_get_property (GObject    *object,
 static void
 on_im_purpose_changed (PosInputSurface *self, GParamSpec *pspec, PosInputMethod *im)
 {
-  GtkWidget *osk_widget = NULL;
+  GtkWidget *widget = NULL;
   PosOskWidgetLayer layer = POS_OSK_WIDGET_LAYER_NORMAL;
   PosInputMethodPurpose purpose;
 
@@ -1099,32 +1118,53 @@ on_im_purpose_changed (PosInputSurface *self, GParamSpec *pspec, PosInputMethod
     break;
   case POS_INPUT_METHOD_PURPOSE_DATE:
   case POS_INPUT_METHOD_PURPOSE_DATETIME:
-  case POS_INPUT_METHOD_PURPOSE_DIGITS:
-  case POS_INPUT_METHOD_PURPOSE_NUMBER:
-  case POS_INPUT_METHOD_PURPOSE_PHONE:
-  case POS_INPUT_METHOD_PURPOSE_PIN:
   case POS_INPUT_METHOD_PURPOSE_TIME:
     layer = POS_OSK_WIDGET_LAYER_SYMBOLS;
     break;
+  case POS_INPUT_METHOD_PURPOSE_NUMBER:
+    pos_keypad_set_decimal_separator_visible (POS_KEYPAD (self->keypad), TRUE);
+    pos_keypad_set_symbols_visible (POS_KEYPAD (self->keypad), FALSE);
+    pos_keypad_set_letters_visible (POS_KEYPAD (self->keypad), FALSE);
+    widget = self->keypad;
+    break;
+  case POS_INPUT_METHOD_PURPOSE_DIGITS:
+    pos_keypad_set_decimal_separator_visible (POS_KEYPAD (self->keypad), FALSE);
+    pos_keypad_set_symbols_visible (POS_KEYPAD (self->keypad), FALSE);
+    pos_keypad_set_letters_visible (POS_KEYPAD (self->keypad), FALSE);
+    widget = self->keypad;
+    break;
+  case POS_INPUT_METHOD_PURPOSE_PIN:
+    pos_keypad_set_decimal_separator_visible (POS_KEYPAD (self->keypad), FALSE);
+    pos_keypad_set_symbols_visible (POS_KEYPAD (self->keypad), FALSE);
+    pos_keypad_set_letters_visible (POS_KEYPAD (self->keypad), TRUE);
+    widget = self->keypad;
+    break;
+  case POS_INPUT_METHOD_PURPOSE_PHONE:
+    pos_keypad_set_decimal_separator_visible (POS_KEYPAD (self->keypad), FALSE);
+    pos_keypad_set_symbols_visible (POS_KEYPAD (self->keypad), TRUE);
+    pos_keypad_set_letters_visible (POS_KEYPAD (self->keypad), TRUE);
+    widget = self->keypad;
+    break;
   case POS_INPUT_METHOD_PURPOSE_TERMINAL:
-    osk_widget = self->osk_terminal;
+    widget = self->osk_terminal;
     break;
   default:
     g_return_if_reached ();
   }
 
-  if (osk_widget == NULL) {
-    osk_widget = hdy_deck_get_visible_child (self->deck);
+  if (widget == NULL) {
+    widget = hdy_deck_get_visible_child (self->deck);
     /* If no "special" layout, Switch back to the last language layer */
-    if (!POS_INPUT_SURFACE_IS_LANG_LAYOUT (osk_widget))
-      osk_widget = self->last_layout;
+    if (!POS_INPUT_SURFACE_IS_LANG_LAYOUT (widget))
+      widget = self->last_layout;
+  } else if (POS_IS_OSK_WIDGET (widget)) {
+    pos_osk_widget_set_layer (POS_OSK_WIDGET (widget), layer);
+    g_debug ("Layout: %s, purpose: %d",
+             pos_osk_widget_get_name (POS_OSK_WIDGET (widget)),
+             purpose);
   }
-  g_debug ("Layout: %s, purpose: %d",
-           pos_osk_widget_get_name (POS_OSK_WIDGET (osk_widget)),
-           purpose);
-  hdy_deck_set_visible_child (self->deck, osk_widget);
 
-  pos_osk_widget_set_layer (POS_OSK_WIDGET (osk_widget), layer);
+  hdy_deck_set_visible_child (self->deck, widget);
 }
 
 
@@ -1291,12 +1331,11 @@ pos_input_surface_finalize (GObject *object)
   g_clear_object (&self->input_settings);
   g_clear_object (&self->osk_settings);
   g_clear_object (&self->xkbinfo);
-  g_clear_object (&self->css_provider);
   g_clear_object (&self->clipboard_manager);
   g_clear_object (&self->completer);
   g_clear_object (&self->completer_manager);
   g_clear_object (&self->swipe_down);
-  g_clear_pointer (&self->theme_name, g_free);
+  g_clear_object (&self->style_manager);
   g_clear_pointer (&self->osks, g_hash_table_destroy);
 
   G_OBJECT_CLASS (pos_input_surface_parent_class)->finalize (object);
@@ -1460,14 +1499,16 @@ pos_input_surface_class_init (PosInputSurfaceClass *klass)
 
   g_type_ensure (POS_TYPE_COMPLETION_BAR);
   g_type_ensure (POS_TYPE_EMOJI_PICKER);
+  g_type_ensure (POS_TYPE_KEYPAD);
   g_type_ensure (POS_TYPE_OSK_WIDGET);
   g_type_ensure (POS_TYPE_SHORTCUTS_BAR);
 
   gtk_widget_class_set_template_from_resource (widget_class,
-                                               "/sm/puri/phosh/osk-stub/ui/input-surface.ui");
+                                               "/mobi/phosh/osk-stub/ui/input-surface.ui");
   gtk_widget_class_bind_template_child (widget_class, PosInputSurface, completion_bar);
   gtk_widget_class_bind_template_child (widget_class, PosInputSurface, deck);
   gtk_widget_class_bind_template_child (widget_class, PosInputSurface, emoji_picker);
+  gtk_widget_class_bind_template_child (widget_class, PosInputSurface, keypad);
   gtk_widget_class_bind_template_child (widget_class, PosInputSurface, menu_box_layouts);
   gtk_widget_class_bind_template_child (widget_class, PosInputSurface, menu_popup);
   gtk_widget_class_bind_template_child (widget_class, PosInputSurface, osk_terminal);
@@ -1477,6 +1518,11 @@ pos_input_surface_class_init (PosInputSurfaceClass *klass)
   gtk_widget_class_bind_template_callback (widget_class, on_emoji_picked);
   gtk_widget_class_bind_template_callback (widget_class, on_emoji_picker_done);
   gtk_widget_class_bind_template_callback (widget_class, on_emoji_picker_delete_last);
+
+  gtk_widget_class_bind_template_callback (widget_class, on_keypad_symbol_pressed);
+  gtk_widget_class_bind_template_callback (widget_class, on_keypad_done);
+
+  gtk_widget_class_bind_template_callback (widget_class, on_latched_modifiers_changed);
   gtk_widget_class_bind_template_callback (widget_class, on_num_shortcuts_changed);
   gtk_widget_class_bind_template_callback (widget_class, on_osk_key_down);
   gtk_widget_class_bind_template_callback (widget_class, on_osk_key_symbol);
@@ -1502,10 +1548,10 @@ pos_input_surface_class_init (PosInputSurfaceClass *klass)
    * A completer implementing the #PosCompleter interface.
    *
    */
-   props[PROP_COMPLETER] =
-     g_param_spec_object ("completer", "", "",
-                          POS_TYPE_COMPLETER,
-                          G_PARAM_READABLE | G_PARAM_EXPLICIT_NOTIFY | G_PARAM_STATIC_STRINGS);
+  props[PROP_COMPLETER] =
+    g_param_spec_object ("completer", "", "",
+                         POS_TYPE_COMPLETER,
+                         G_PARAM_READABLE | G_PARAM_EXPLICIT_NOTIFY | G_PARAM_STATIC_STRINGS);
   /**
    * PosInputSurface:completer-manager:
    *
@@ -1819,7 +1865,6 @@ on_completion_mode_changed (PosInputSurface *self, const char *key, GSettings *s
 
 static GActionEntry entries[] =
 {
-  { .name = "clipboard-copy", .activate = clipboard_copy_activated },
   { .name = "clipboard-paste", .activate = clipboard_paste_activated },
   { .name = "settings", .activate = settings_activated },
   { .name = "select-layout", .parameter_type = "s", .state = "\"terminal\"",
@@ -1831,13 +1876,13 @@ static GActionEntry entries[] =
 static void
 pos_input_surface_init (PosInputSurface *self)
 {
-  GtkSettings *gtk_settings;
   g_autoptr (GPropertyAction) completion_action = NULL;
   g_autoptr (GError) err = NULL;
   GAction *action;
 
   gtk_widget_init_template (GTK_WIDGET (self));
 
+  self->style_manager = pos_style_manager_new ();
   self->action_map = g_simple_action_group_new ();
   g_action_map_add_action_entries (G_ACTION_MAP (self->action_map),
                                    entries,
@@ -1879,13 +1924,6 @@ pos_input_surface_init (PosInputSurface *self)
     g_warning ("Failed to set terminal layout: %s", err->message);
   }
 
-  gtk_settings = gtk_settings_get_default ();
-  g_object_set (G_OBJECT (gtk_settings), "gtk-application-prefer-dark-theme", TRUE, NULL);
-
-  g_signal_connect_swapped (gtk_settings, "notify::gtk-theme-name",
-                            G_CALLBACK (on_gtk_theme_name_changed), self);
-  on_gtk_theme_name_changed (self, NULL, gtk_settings);
-
   self->clicked_id = g_signal_add_emission_hook (g_signal_lookup ("clicked", GTK_TYPE_BUTTON), 0,
                                                  on_click_hook, self, NULL);
 
